# 1 "<stdin>"
# 1 "<built-in>"
# 1 "<command line>"
# 1 "<stdin>"
program Forward

  use sep
  use tomo_mod

  implicit none

  integer :: nt, vel_nz, vel_nx, dev_numin, sou_z, rec_z, i, it, ix
  integer :: sou_nx, sou_ox, sou_dx, rec_nx, rec_ox, rec_dx
  real :: dt, vel_dz, vel_oz, vel_dx, vel_ox

  real :: rho
  real, dimension(:), allocatable :: wavelet
  real, dimension(:,:), allocatable :: slow0, wavelet2, data, rf

  integer, dimension(8) :: timer0, timer1
  real :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", vel_nz)
  call from_history("o1", vel_oz)
  call from_history("d1", vel_dz)
  call from_history("n2", vel_nx)
  call from_history("o2", vel_ox)
  call from_history("d2", vel_dx)

  ! Read wavelet information
  call from_aux("wavfile", "d1", dt)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  ! Read recording time
  call from_param("nt", nt)

  ! Read in Density
  call from_param('rho', rho, 1000.0)

  call from_param("dev_numin",dev_numin)

  if(sou_nx .eq. 0) sou_nx=vel_nx
  if(rec_nx .eq. 0) rec_nx=vel_nx


  ! Allocation
  allocate( wavelet(nt))
  allocate( slow0(vel_nz, vel_nx), wavelet2(nt, vel_nx), data(nt, rec_nx))
  allocate( rf(vel_nz, vel_nx))

  call to_history("n1", nt)
  call to_history("o1", 0.0)
  call to_history("d1", dt)
  call to_history("n2", rec_nx)
  call to_history("o2", vel_dx*rec_ox+vel_ox)
  call to_history("d2", vel_dx)
  call to_history("n3", sou_nx)
  call to_history("o3", vel_dx*sou_ox+vel_ox)
  call to_history("d3", vel_dx*sou_dx)
  call to_history("label1", "Time(s)")
  call to_history("label2", "Location(km)")
  call to_history("label3", " ")

  ! Read in the background velocity
  call sep_read(slow0)

  call sep_read(rf, "refl")

  ! Read in the source wavefields
  call sep_read(wavelet, "wavfile")

  ! Scale the source wavefields by rho
  wavelet = wavelet * rho
  call SecDer1d(wavelet, wavelet)

  rf = rf*slow0

  slow0 = dt*dt/slow0/slow0

  call c_init(nt, dt, vel_nz, vel_nx, vel_dz, vel_dx, rec_nx, dev_numin)

  call c_vel_update(slow0)

  do i=sou_ox, (sou_nx-1)*sou_dx+sou_ox, sou_dx

    wavelet2 = 0.
    wavelet2(:,i+1) = wavelet(:)
    data=0.

    call c_model2(sou_z, wavelet2, rf, rec_z, rec_ox, data)

    call sep_write(data)

  end do

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
