# 1 "<stdin>"
# 1 "<built-in>"
# 1 "<command line>"
# 1 "<stdin>"
program Forward

  use sep
  use tomo_mod

  implicit none

  integer :: nt, vel_nz, vel_nx, wfld_j, dev_numin, sou_z, rec_z, ishot, it, ix, iz
  integer :: sou_nx, sou_ox, sou_dx, rec_nx, rec_ox, rec_dx, h_nx, h_ox, ih, ihx, j, tl
  real :: dt, vel_dz, vel_oz, vel_dx, vel_ox, wt

  integer :: n1s, n2s, n3s
  integer :: crop1, crop2

  real :: rho
  real, dimension(:), allocatable :: wavelet
  real, dimension(:,:), allocatable :: slow0, vel_sq, injected, deltaS
  real, dimension(:,:,:), allocatable :: sou_wfld, rec_wfld, deltaI, data_obs
  real, dimension(:,:,:), allocatable :: sou_wfld2, rec_wfld2

  integer, dimension(8) :: timer0, timer1
  real :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", vel_nz)
  call from_history("o1", vel_oz)
  call from_history("d1", vel_dz)
  call from_history("n2", vel_nx)
  call from_history("o2", vel_ox)
  call from_history("d2", vel_dx)

  ! Read wavelet information
  call from_aux("wavfile", "d1", dt)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  ! Read recording time
  call from_param("nt", nt)

  ! Read in Density
  call from_param('rho', rho, 1000.0)

  call from_param("h_nx",h_nx)

  if(mod(h_nx,2) .ne. 1) h_nx = h_nx + 1
  h_ox = -1*(h_nx-1)/2

  call from_param("wfld_j", wfld_j)
  call from_param("dev_numin", dev_numin, 0)
  call from_param("crop1", crop1)
  call from_param("crop2", crop2)

  n3s = ceiling(1.0*nt/wfld_j)
  n1s = vel_nz-2*crop1
  n2s = vel_nx-2*crop2

  if(sou_nx .eq. 0) sou_nx=vel_nx
  if(rec_nx .eq. 0) rec_nx=vel_nx


  ! Allocation
  allocate( wavelet(nt))
  allocate( slow0(vel_nz, vel_nx), vel_sq(vel_nz, vel_nx), injected(nt, vel_nx), data_obs(nt, rec_nx, sou_nx))
  allocate( deltaS(n1s, n2s), deltaI(n1s, n2s, h_nx))
  allocate( sou_wfld(n1s, n2s, n3s), rec_wfld(n1s, n2s, n3s))
  allocate( sou_wfld2(n1s, n2s, n3s), rec_wfld2(n1s, n2s, n3s))

  call to_history("n1", n1s)
  call to_history("o1", 0.)
  call to_history("d1", vel_dz)
  call to_history("n2", n2s)
  call to_history("o2", 0.)
  call to_history("d2", vel_dx)
  call to_history("n3", 1)
  call to_history("o3", 0.)
  call to_history("d3", 1.)

  ! Read in the background velocity
  call sep_read(slow0)

  call sep_read(deltaI, "deltaI")

  call sep_read(data_obs, "data_obs")

  ! Read in the source wavefields
  call sep_read(wavelet, "wavfile")

  ! Scale the source wavefields by rho
  wavelet = wavelet * rho
  call SecDer1d(wavelet, wavelet)

  tl=rec_nx/10

  vel_sq = dt*dt/slow0/slow0

  call c_init(nt, dt, vel_nz, vel_nx, vel_dz, vel_dx, rec_nx, rec_ox, n1s, n2s, h_nx, h_ox, crop1, crop2, wfld_j, dev_numin)

  call c_vel_update(vel_sq)

  deltaS = 0.

  do ishot=1,sou_nx

    injected = 0.
    injected(:,sou_ox+1+(ishot-1)*sou_dx) = wavelet(:)

    do j=1,tl
      wt=1.*j/tl
      injected(:,j+crop2) = injected(:,j+crop2)*wt
    end do

    do j=rec_nx-tl+1,rec_nx
      wt=(rec_nx-j+1.)/tl
      injected(:,j+crop2) = injected(:,j+crop2)*wt
    end do

    sou_wfld = 0.
    sou_wfld2 = 0.
    call c_prop4(sou_z, injected, deltaI, sou_wfld, sou_wfld2, 0)

    injected = 0.
    do ix=1,rec_nx
      injected(:,rec_ox+ix) = data_obs(:,ix,ishot)
    end do

    do j=1,tl
      wt=1.*j/tl
      injected(:,j+crop2) = injected(:,j+crop2)*wt
    end do

    do j=rec_nx-tl+1,rec_nx
      wt=(rec_nx-j+1.)/tl
      injected(:,j+crop2) = injected(:,j+crop2)*wt
    end do

    rec_wfld = 0.
    rec_wfld2 = 0.
    call c_prop4(rec_z, injected, deltaI, rec_wfld, rec_wfld2, n3s-1)

    deltaS(:,:) = deltas(:,:) + sum(sou_wfld(:,:,:)*rec_wfld2(:,:,:)+sou_wfld2(:,:,:)*rec_wfld(:,:,:),3)*slow0(1+crop1:n1s+crop1,1+crop2:n2s+crop2)

  end do

  call sep_write(deltaS)

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
