# 1 "<stdin>"
# 1 "<built-in>"
# 1 "<command line>"
# 1 "<stdin>"
module tomo_mod

implicit none

contains

subroutine FirstDer(souIn, souOut, n1)
  !! take 1st derivative
  integer :: i, n1
  real, dimension(:) :: souIn, souOut
  real, dimension(n1) :: temp

  temp = souIn

  do i=1,n1-1
    souOut(i) = temp(i+1) - temp(i)
  end do

  souOut(n1) = souOut(n1-1)

end subroutine

!!----------------------------------------------------------

subroutine SecDer1d(souIn, souOut)
  !! take 2nd derivative
  integer :: i
  real, dimension(:) :: souIn, souOut
  real, dimension(size(souIn)) :: temp

  temp = souIn
  do i=2,size(souIn)-1
    souOut(i) = 2*temp(i) - temp(i+1) - temp(i-1)
  end do

  souOut(1) = souOut(2)
  souOut(size(souIn)) = souOut(size(souIn)-1)

end subroutine

!!----------------------------------------------------------

subroutine SecDer2d(souIn, souOut)
  !! take 2nd derivative on first axis
  integer :: i
  real, dimension(:,:) :: souIn, souOut
  real, dimension(size(souIn,1),size(souIn,2)) :: temp

  temp = souIn
  do i=2,size(souIn,1)-1
    souOut(i,:) = 2*temp(i,:) - temp(i+1,:) - temp(i-1,:)
  end do

  souOut(1,:) = souOut(2,:)
  souOut(size(souIn,1),:) = souOut(size(souIn,1)-1,:)

end subroutine

!!----------------------------------------------------------

subroutine SecDer3d(souIn, souOut)
  !! take 2nd derivative on third axis
  integer :: i
  real, dimension(:,:,:) :: souIn, souOut
  real, dimension(size(souIn,1),size(souIn,2),size(souIn,3)) :: temp

  temp = souIn
  do i=2,size(souIn,3)-1
    souOut(:,:,i) = 2*temp(:,:,i) - temp(:,:,i+1) - temp(:,:,i-1)
  end do

  souOut(:,:,1) = souOut(:,:,2)
  souOut(:,:,size(souIn,3)) = souOut(:,:,size(souIn,3)-1)

end subroutine

!!--------------------------------------------------------------------------
!! 2D max cross correlation

  subroutine MaxCrossCor(v1, v2, n1, n2, window, lagmax)

    integer :: n1, n2, i, j, k, lag, window
    integer, dimension(:) :: lagmax
    real :: coeff, coeffmax
    real, dimension(:,:) :: v1, v2

    do k=1,n2
      coeffmax=0.
      do j=1,window+1
        lag = j - window - 1
        coeff = 0.
        do i=1,n1+lag
          coeff = coeff + v1(i-lag,k)*v2(i,k)
        end do
        if(coeffmax<coeff) then
          coeffmax = coeff
          lagmax(k) = lag
        end if
      end do
      do j=window+2,2*window+1
        lag = j - window - 1
        coeff = 0.
        do i=1+lag,n1
          coeff = coeff + v1(i-lag,k)*v2(i,k)
        end do
        if(coeffmax<coeff) then
          coeffmax = coeff
          lagmax(k) = lag
        end if
      end do
    end do

  end subroutine

!!--------------------------------------------------------------------------
!! 2D cross correlation

  subroutine CrossCorOut(v1, v2, n1, n2, window, xcorr)

    integer :: n1, n2, i, j, k, lag, window
    real, dimension(:,:) :: v1, v2, xcorr

    xcorr = 0.

    do k=1,n2
      do j=1,window+1
        lag = j - window - 1
        do i=1,n1+lag
          xcorr(j,k) = xcorr(j,k) + v1(i-lag,k)*v2(i,k)
        end do
      end do
      do j=window+2,2*window+1
        lag = j - window - 1
        do i=1+lag,n1
          xcorr(j,k) = xcorr(j,k) + v1(i-lag,k)*v2(i,k)
        end do
      end do
    end do

  end subroutine

end module
