module angle_mod

  implicit none

  integer, private                  :: nx, nh, nz
  real, private                     :: dp, dh, dz
  real, private                     :: op, oh, oz

  contains

  subroutine angle_init(nz_in, oz_in, dz_in, nx_in, op_in, dp_in, nh_in, oh_in, dh_in)

    integer             :: nz_in, nx_in, nh_in
    real                :: oz_in, dz_in, op_in, dp_in, oh_in, dh_in

    nz = nz_in
    oz = oz_in
    dz = dz_in
    nx = nx_in
    op = op_in
    dp = dp_in
    nh = nh_in
    oh = oh_in
    dh = dh_in

  end subroutine

  subroutine angle_op(adj, angle, offset)

    integer                           :: ix, ip, ih, iz
    real                              :: x, p, h, z, t
    integer                           :: itl, itr
    real                              :: cl, cr, tt
    logical                           :: adj

    real, dimension(:,:,:)            :: offset, angle


    if (adj) then
      angle(iz,ix,ip) = 0.0
    else
      offset(itl,ix,ih) = 0.0
    end if

    do ix = 1,nx
      do ip = 1,nh
        p = op + (ip-1) * dp
        do ih = 1,nh
          h = oh + (ih-1) * dh
          do iz = 1,nz
            z = oz + (iz-1) * dz
            t  = z + tan(p*3.14159265/180.0)*h

            tt  = 1.0 + (t-oz)/dz
            itr = ceiling(tt)
            itl = floor(tt)
            cr = tt - itl
            cl = 1.0 - cr
            if( itr >=1 .and. itr <= nz) then
              if (adj) then
                angle(iz,ix,ip) = angle(iz,ix,ip) + cl*offset(itl,ix,ih)+cr*offset(itr,ix,ih)
              else
                offset(itl,ix,ih) = offset(itl,ix,ih) + cl*angle(iz,ix,ip)
                offset(itr,ix,ih) = offset(itr,ix,ih) + cr*angle(iz,ix,ip)
              end if
            end if
          end do
        end do
      end do
    end do

  end subroutine

end module 
