module combine_mod

implicit none

integer, private :: n1,n2

contains

subroutine step_init(n1_in, n2_in)
  integer            :: n1_in, n2_in
  n1 = n1_in
  n2 = n2_in  
end subroutine

!---------------------------------------------------------------------------

subroutine step1(vtrue, v0, dv, a)
  real,dimension(:,:)  :: vtrue, v0, dv
  real                 :: a
  call step12(vtrue-v0, dv, a)
end subroutine

!---------------------------------------------------------------------------

subroutine step12(dvtrue, dv, a)
  real, dimension(n1*n2)       :: dvtrue, dv
  real                         :: a
  a = dot_product(dvtrue,dv)/dot_product(dv,dv)
end subroutine

!---------------------------------------------------------------------------

subroutine step2(vtrue, v0, dv1, dv2, a1, a2)
  real,dimension(:,:)  :: vtrue, v0, dv1, dv2
  real                 :: a1, a2
  call step22(vtrue-v0, dv1, dv2, a1, a2)
end subroutine

!---------------------------------------------------------------------------

subroutine step22(dvtrue, dv1, dv2, a1, a2)
  real, dimension(n1*n2)       :: dvtrue, dv1, dv2
  real, dimension(2,2)         :: G, Gi
  real                         :: a1, a2, t1, t2
  G(1,1) = dot_product(dv1,dv1)
  G(1,2) = dot_product(dv1,dv2)
  G(2,1) = G(1,2)
  G(2,2) = dot_product(dv2,dv2)
  Gi(1,1) = G(2,2)
  Gi(1,2) = -1*G(1,2)
  Gi(2,1) = -1*G(2,1)
  Gi(2,2) = G(1,1)
  Gi = Gi/(G(1,1)*G(2,2)-G(2,1)*G(1,2))
  t1 = dot_product(dv1,dvtrue)
  t2 = dot_product(dv2,dvtrue)
  a1 = Gi(1,1)*t1+Gi(1,2)*t2
  a2 = Gi(2,1)*t1+Gi(2,2)*t2
end subroutine

!---------------------------------------------------------------------------

subroutine step3(vtrue, v0, dv1, dv2, dv3, a1, a2, a3)
  real,dimension(:,:)  :: vtrue, v0, dv1, dv2, dv3
  real                 :: a1, a2, a3
  call step33(vtrue-v0, dv1, dv2, dv3, a1, a2, a3)
end subroutine

!---------------------------------------------------------------------------

subroutine step33(dvtrue, dv1, dv2, dv3, a1, a2, a3)
  real, dimension(n1*n2)       :: dvtrue, dv1, dv2, dv3
  real, dimension(3,3)         :: Gi
  real                         :: v11, v12, v13, v22, v23, v33
  real                         :: a1, a2, a3, t1, t2, t3, det
  v11 = dot_product(dv1,dv1)
  v12 = dot_product(dv1,dv2)
  v13 = dot_product(dv1,dv3)
  v22 = dot_product(dv2,dv2)
  v23 = dot_product(dv2,dv3)
  v33 = dot_product(dv3,dv3)

  det = (v33*v12**2 - 2*v12*v13*v23 + v22*v13**2 + v11*v23**2 - v11*v22*v33)

  Gi(1,1) = -(v22*v33 - v23**2)/det
  Gi(2,1) = -(v13*v23 - v12*v33)/det
  Gi(3,1) = -(v12*v23 - v13*v22)/det
  Gi(1,2) = Gi(2,1)
  Gi(2,2) = -(v11*v33 - v13**2)/det
  Gi(3,2) = -(v12*v13 - v11*v23)/det
  Gi(1,3) = Gi(3,1)
  Gi(2,3) = Gi(3,2)
  Gi(3,3) = -(v11*v22 - v12**2)/det

  t1 = dot_product(dv1,dvtrue)
  t2 = dot_product(dv2,dvtrue)
  t2 = dot_product(dv3,dvtrue)

  a1 = Gi(1,1)*t1+Gi(1,2)*t2+Gi(1,3)*t3
  a2 = Gi(2,1)*t1+Gi(2,2)*t2+Gi(2,3)*t3
  a3 = Gi(3,1)*t1+Gi(3,2)*t2+Gi(3,3)*t3

end subroutine

end module
