#include "cudaErrors.cu"

#define dloc(z,x) (z)+(x)*n1gpu
#define min2(v1,v2) (((v1)<(v2))?(v1):(v2)) 

__constant__ int nt_gpu;
__constant__ float dt_gpu;
__constant__ int n1gpu;
__constant__ int n2gpu;
__constant__ int nrec_gpu;
__constant__ int orec_gpu;
__constant__ int crop1_gpu;
__constant__ int crop2_gpu;
__constant__ int n1s_gpu;
__constant__ int n2s_gpu;
__constant__ int wfld_j_gpu;
texture <float, 2, cudaReadModeElementType> vel_ref;
cudaChannelFormatDesc cf;
cudaArray *vel_array=0;
__device__ int min3(int v1,int v2,int v3){return min2(min2(v1,v2),v3);}
__device__ int min4(int v1,int v2,int v3, int v4){return min2(min2(v1,v2),min2(v3,v4));}

float *p0_gpu, *p1_gpu, *p2_gpu, *p3_gpu, *source_gpu, *rf_gpu, *wfld1_gpu, *wfld2_gpu;

#include "wave_kernel2d8o.cu"

__global__ void inject_kernel(int z, int it, float *source_gpu, float *p)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    if(i>=n2gpu) return;
    p[dloc(z,i)] += source_gpu[it+i*nt_gpu];
}


__global__ void inject2_kernel(float *p1, float *rf_gpu, float *p3)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    register int j=blockIdx.y*blockDim.y+threadIdx.y;
    p3[dloc(i,j)] += p1[dloc(i,j)]*rf_gpu[dloc(i,j)];
}


__global__ void collect2_kernel(float *wfld_gpu, float *p)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    register int j=blockIdx.y*blockDim.y+threadIdx.y;
    if(i>=n1s_gpu || j>=n2s_gpu) return;
    wfld_gpu[i+j*n1s_gpu] = p[dloc(i+crop1_gpu,j+crop2_gpu)];
}


extern "C" void gpu_init(int *nt, float *dt, int *n1, int *n2, float *d1, float *d2, int *nrec, int *orec, int *n1s, int *n2s, int *crop1, int *crop2, int *wfld_j, int *gpu_id)
{

    cudaSetDevice(*gpu_id);

    cuda_call(cudaMalloc(    (void**) &p0_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &p1_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &p2_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &p3_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &source_gpu,  (*nt)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &rf_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &wfld1_gpu,  (*n1s)*(*n2s)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &wfld2_gpu,  (*n1s)*(*n2s)*sizeof(float)));

    //bind texture reference to array
    cf = cudaCreateChannelDesc<float>();
    cuda_call(cudaMallocArray(&vel_array, &cf, *n1, *n2));

    cuda_call(cudaMemcpyToSymbol(n1gpu, n1, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n2gpu, n2, sizeof(int)));

    cuda_call(cudaMemcpyToSymbol(nt_gpu,  nt,   sizeof(int)));

    cuda_call(cudaMemcpyToSymbol(nrec_gpu,  nrec,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(orec_gpu,  orec,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(crop1_gpu,  crop1,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(crop2_gpu,  crop2,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n1s_gpu,  n1s,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n2s_gpu,  n2s,   sizeof(int)));

    cuda_call(cudaMemcpyToSymbol(wfld_j_gpu,  wfld_j,   sizeof(int)));

    float coeffs_cpu[COEFFS_SIZE]=get_coeffs(*d1,*d2);
    cuda_call(cudaMemcpyToSymbol(coeffs, coeffs_cpu, COEFFS_SIZE*sizeof(float)));

}

extern "C" void gpu_vel_update(float *cpu_space)
{
    int n1, n2;

    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int))); 
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int))); 
    cuda_call(cudaMemcpyToArray(vel_array, 0,0, cpu_space, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaBindTextureToArray(vel_ref, vel_array));
}

extern "C" void gpu_prop(int *sou_z, float *source_cpu, float *wfld, int *i_start)
{
    int nt, n1, n2, n3, wfld_j;
    int n1s, n2s;
    float *ptemp;

    cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1s, n1s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2s, n2s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int)));

    cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemset(p0_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p1_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(wfld1_gpu, 0,n1s*n2s*sizeof(float)));

    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);

    n3=*i_start;

    if(n3 == 0)
    {
        for (int it=0; it<nt; it++)
        {
            kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
            ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
            kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));

            if(it%wfld_j==0)
            {
                kernel_exec(collect2_kernel<<<dimGrid, dimBlock>>>(wfld1_gpu, p1_gpu));
                cuda_call(cudaMemcpy(&wfld[n1s*n2s*n3], wfld1_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
                n3++;
            }

        }
    }
    else
    {
        for (int it=nt-1; it>=0; it--)
        {
            kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
            ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
            kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));

            if(it%wfld_j==0)
            {
                kernel_exec(collect2_kernel<<<dimGrid, dimBlock>>>(wfld1_gpu, p1_gpu));
                cuda_call(cudaMemcpy(&wfld[n1s*n2s*n3], wfld1_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
                n3--;
            }

        }
    }
}

extern "C" void gpu_prop3(int *sou_z, float *source_cpu, float *rf_cpu, float *wfld, float *wfld2, int *i_start)
{
    int nt, n1, n2, n3, wfld_j;
    int n1s, n2s;
    float *ptemp;

    cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1s, n1s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2s, n2s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int)));

    cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemcpy(rf_gpu, rf_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemset(p0_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p1_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p2_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p3_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(wfld1_gpu, 0,n1s*n2s*sizeof(float)));
    cuda_call(cudaMemset(wfld2_gpu, 0,n1s*n2s*sizeof(float)));

    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);

    n3=*i_start;

//    printf("n3 start = %d \n",n3);

    if(n3 == 0)
    {
        for (int it=0; it<nt; it++)
        {
            kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
            ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
            kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));

            kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p2_gpu, p3_gpu, p2_gpu));
            ptemp=p2_gpu; p2_gpu=p3_gpu; p3_gpu=ptemp;
            kernel_exec(inject2_kernel<<<dimGrid, dimBlock>>>(p1_gpu, rf_gpu, p3_gpu));

            if(it%wfld_j==0)
            {
                kernel_exec(collect2_kernel<<<dimGrid, dimBlock>>>(wfld1_gpu, p1_gpu));
                kernel_exec(collect2_kernel<<<dimGrid, dimBlock>>>(wfld2_gpu, p3_gpu));
                cuda_call(cudaMemcpy(&wfld[n1s*n2s*n3], wfld1_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
                cuda_call(cudaMemcpy(&wfld2[n1s*n2s*n3], wfld2_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
                n3++;
            }

        }
    }
    else
    {
        for (int it=nt-1; it>=0; it--)
        {
            kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
            ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
            kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));

            kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p2_gpu, p3_gpu, p2_gpu));
            ptemp=p2_gpu; p2_gpu=p3_gpu; p3_gpu=ptemp;
            kernel_exec(inject2_kernel<<<dimGrid, dimBlock>>>(p1_gpu, rf_gpu, p3_gpu));

            if(it%wfld_j==0)
            {
                kernel_exec(collect2_kernel<<<dimGrid, dimBlock>>>(wfld1_gpu, p1_gpu));
                kernel_exec(collect2_kernel<<<dimGrid, dimBlock>>>(wfld2_gpu, p3_gpu));
                cuda_call(cudaMemcpy(&wfld[n1s*n2s*n3], wfld1_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
                cuda_call(cudaMemcpy(&wfld2[n1s*n2s*n3], wfld2_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
                n3--;
            }

        }
    }
}
